///////////////////////////////////////////////////////////////////////////////
// Copyright (C) 2009 Sebastian Pancratz                                     //
//                                                                           //
// Distributed under the terms of the GNU General Public License as          //
// published by the Free Software Foundation; either version 2 of the        //
// License, or (at your option) any later version.                           //
//                                                                           //
// http://www.gnu.org/licenses/                                              //
///////////////////////////////////////////////////////////////////////////////

/**
 * \file    monomial.c
 * \brief   Fast implementation of monomials for multivariate polynomial rings
 * \author  Sebastian Pancratz
 * \date    Oct--Nov 2009
 * 
 * All methods support aliasing of their inputs and outputs \e unless
 * explicitly stated otherwise.
 *
 * \todo  Implement further monomial orderings.
 */

/**
 * \mainpage
 *
 * \author Sebastian Pancratz
 * \date   Oct--Nov 2009
 *
 * \par Acknowledgements:
 * I would like to thank Andrew Lewis from the Computer Laboratory at the 
 * University of Cambridge for helpful discussions.
 * 
 * \section Overview
 * 
 * The aim of this work is to provide a basic but very thin and fast 
 * implementation of monomials with only minimal overhead.  The main step 
 * in order to achieve this is encoding a given monomial in one machine 
 * word, or possibly two, although this case is handled transparently. As 
 * a consequence of this design decision, this implementation can only be 
 * used for monomials in a fixed small number of variables and with small 
 * non-negative exponents.
 * 
 * \section Introduction  Detailed introduction
 * 
 * We assume that a number \f$n+1\f$ of variables is fixed from the beginning, 
 * where \f$1 \leq n \leq 7\f$, and that we only consider monomials with 
 * non-negative partial degrees of at most \f$2^8 - 1\f$.  Using 8 bits per 
 * exponent, we can encode the monomial \f$i = (i_0, \dotsc, i_n)\f$ as 
 * \f[
 * \sigma(i) = \sum_{j=0}^n i_j 2^{j 8}
 * \f]
 * in a 64-bit word.
 * 
 * It turns out that, for many operations on monomials, this form is very 
 * convenient in the sense that the operation can be performed on 
 * \f$\sigma(i)\f$ as one word instead of \f$i_0, \dotsc, i_n\f$ involving 
 * \f$n+1\f$ words.  Examples of this are multiplication and division, 
 * which correspond to addition and subtraction, and the inverse 
 * lexicographical order, which corresponds to the operator \c <.  As an 
 * aside, the problem of possible overflow is burried in the our assumption 
 * that no monomial operation will involve partial exponents exceeding 
 * \f$2^8 - 1\f$.  Other operations, for example, checking whether one 
 * monomial is divisible by another, involve explicitly considering partial 
 * exponents using bitmasks.
 * 
 * \section Orderings  Monomial orderings
 * 
 * The following (global) monomial orderings are typically considered:
 * 
 * - Lexicographic ordering (\em lex).
 *   \f[i < i' \iff \exists j \quad i_0 = i_0', \dotsc, i_{j-1} = i_{j-1}', i_j < i_j'\f]
 * 
 * - Inverse lexicographic ordering (\em invlex).
 *   \f[i < i' \iff \exists j \quad i_n = i_n', \dotsc, i_{j+1} = i_{j+1}', i_j < i_j'\f]
 * 
 * - Degree lexicographic (\em deglex).
 *   \f[i < i' \iff \deg(i) < \deg(i') \text{ or } \bigl(\deg(i) < \deg(i') \text{ and } \exists j \quad i_0 = i_0', \dotsc, i_{j-1} = i_{j-1}, i_j < i_j'\bigr)\f]
 * 
 * - Degree reverse lexicographic (\em degrevlex).
 *   \f[i < i' \iff \deg(i) < \deg(i') \text{ or } \bigl(\deg(i) < \deg(i') \text{ and } \exists j \quad i_n = i_n', \dotsc, i_{j+1} = i_{j+1}, i_j > i_j'\bigr)\f]
 * 
 * In this context, it is very useful to note that, in the inverse 
 * lexicographic ordering, we have \f$i < i'\f$ if and only 
 * \f$\sigma(i) < \sigma(i')\f$.  While the first test could involve \f$n+1\f$ 
 * word comparisons plus the overhead of a loop, the second test only involves 
 * one word comparison.
 * 
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <inttypes.h>

#include "monomial.h"

///////////////////////////////////////////////////////////////////////////////
// Macros                                                                    //
///////////////////////////////////////////////////////////////////////////////

/**
 * \def _MON_EXP_LENGTH_
 *
 * Number of digits (in base 10) for the exponent \c e.
 * 
 * \warning  This macro possibly evaluates its arguments multiple times.
 */
#define _MON_EXP_LENGTH_(e)  ((e) < 10 ? 1 : (e) < 100 ? 2 : (e) < 1000 ? 3 : (e) < 10000 ? 4 : (e) < 100000 ? 5 : (e) < 1000000 ? 6 : (e) < 10000000 ? 7 : (e) < 100000000 ? 8 : (e) < 1000000000 ? 9 : 10)

///////////////////////////////////////////////////////////////////////////////
// Helper functions                                                          //
///////////////////////////////////////////////////////////////////////////////

/**
 * Returns the binomial coefficient \f$\binom{n}{k}\f$.
 *
 * Special cases are handled as follows.  If \f$n < k\f$, returns \f$0\f$. 
 * If \f$k = 0\f$ or \f$k = n\f$, returns \f$1\f$.
 */
uint32_t _binom_(uint32_t n, uint32_t k)
{
    uint32_t i, d, res;
    if (n < k)
        return (uint32_t) 0;
    if (k > (n >> 1))
        k = n - k;
    if (k == 0)
        return (uint32_t) 1;
    d = n - k;
    res = d + 1;
    for (i = 2; i <= k; i++)
    {
        res *= d + i;
        res /= i;
    }
    return res;
}

///////////////////////////////////////////////////////////////////////////////
// Implementations of methods declared in the header                         //
///////////////////////////////////////////////////////////////////////////////

/**
 * Returns the total degree of the monomial \c op.
 *
 * Note that, since we assume the individual degrees to be less than 256 and 
 * support at most eight variables, the result fits comfortably in a 32-bit 
 * unsigned integer.
 */
uint32_t mon_degree(monomial_t op)
{
	uint32_t result = op & MON_BITMASK_BLOCK;
	while (op != 0)
		result += (op >>= MON_BITS_PER_EXP) & MON_BITMASK_BLOCK;
	return result;
}

/**
 * Returns whether the monomial \c x divides the monomial \c y.
 *
 * The return value will be either \c 1 or \c 0, depending on whether or not 
 * \c x divides \c y.
 */
int32_t mon_divides(monomial_t x, monomial_t y)
{
    while (x != 0)
    {
        if ((x & MON_BITMASK_BLOCK) > (y & MON_BITMASK_BLOCK))
            return 0;
        x >>= MON_BITS_PER_EXP;
        y >>= MON_BITS_PER_EXP;
    }
    return 1;
}

/**
 * Returns a string representation of the monomial \c x in \f$n+1\f$ variables.
 */
char * mon_to_string(monomial_t x, const uint8_t n)
{
    char * str;
    exponent_t e;
    int8_t i;
    size_t off;
    
    str = (char *) malloc((3 + (n + 1) * 4 + 1) * sizeof(char));
    if (str == NULL)
    {
        printf("ERROR (mon_to_string).  Memory allocation failed.\n");
        exit(-1);
    }
    
    sprintf(str, "%d ", n + 1);
    for (i = 0; i <= n; i++)
    {
        e = x & MON_BITMASK_BLOCK;
        off = strlen(str);
        sprintf(str + off, " %" PRIu8, e);
        x >>= MON_BITS_PER_EXP;
    }
    
    return str;
}

/**
 * Returns a pretty string representation of the monomial \c x in \f$n+1\f$ 
 * variables.
 *
 * Since \c x is a monomial in <tt>n + 1</tt> variables, the array of 
 * <tt>char</tt>s \c vars should have length at least that, so that the 
 * <tt>i</tt>th entry can be used as a symbol for the <tt>i</tt>th variable 
 * in \c x.
 * 
 * Alternatively, if <tt>vars == NULL</tt>, an array of default variable 
 * names is used.
 */
char * mon_to_string_pretty(monomial_t x, const uint8_t n, const char * vars)
{
    char * str;
    exponent_t e;
    size_t off;
    int8_t i;
    
    if (MON_IS_ONE(x))
    {
        str = (char *) malloc(2 * sizeof(char));
        if (str == NULL)
        {
            printf("ERROR (mon_to_string_pretty).  Memory allocation failed.\n");
            exit(-1);
        }
        
        str[0] = '1';
        str[1] = '\0';
        return str;
    }
    
    str = (char *) malloc((n + 1) * 6 * sizeof(char));
    if (str == NULL)
    {
        printf("ERROR (mon_to_string_pretty).  Memory allocation failed.\n");
        exit(-1);
    }
    
    off = 0;
    i = 0;
    while (x)
    {
        e = x & MON_BITMASK_BLOCK;
        
        if (e)
        {
            if (e == 1)
            {
                str[off++] = vars ? vars[i] : MON_VARS(n)[i];
                str[off++] = ' ';
            }
            else
            {
                str[off++] = vars ? vars[i] : MON_VARS(n)[i];
                str[off++] = '^';
                sprintf(str + off, "%" PRIu8, e);
                off += _MON_EXP_LENGTH_(e);
                str[off++] = ' ';
            }
        }
        
        x >>= MON_BITS_PER_EXP;
        i++;
    }
    
    str[off-1] = '\0';
    return str;
}

/**
 * Returns the monomial given by the string representation \c str.
 */
monomial_t mon_from_string(char * str)
{
    uint8_t i, n;    // Number of variables, *not* this minus one
    monomial_t rop;  // Output
    exponent_t e;    // Temp
    size_t off;
    
    n = atoi(str);
    if (2 > n | n > 8)
    {
        printf("ERROR (mon_from_string).  Bad number of variables.\n");
        exit(-1);
    }
    
    MON_INIT(rop);
    off = 3;
    for (i = 0; i < n; i++)
    {
        e = (exponent_t) atoi(str + off);
        MON_SET_EXP(rop, i, e);
        off += _MON_EXP_LENGTH_(e) + 1;
    }
    
    return rop;
}

/**
 * Generates all monomials in <tt>n + 1</tt> variables of total degree \c d.
 *
 * This method returns an array of length \f$\binom{n+d}{d}\f$, which is also 
 * the value of <tt>*len</tt> on exit.  The monomials are arrange in 
 * descending lexicographical order.
 */
monomial_t * mon_generate_by_degree(uint32_t * len, const uint8_t n, const uint32_t d)
{
    monomial_t * st;  // Stack (from the bottom), list (from the top)
    uint32_t * sums;
    uint32_t * inds;
    int32_t i;        // Pointer to the top of the stack
    int32_t j;        // Pointer to the bottom of the list
    uint32_t N;       // $\binom{n+d}{d}$
    monomial_t m;
    uint32_t sum, ind;
    exponent_t exp;
    
    // Special case d == 0
    if (d == 0)
    {
        st = (monomial_t *) malloc(sizeof(monomial_t));
        if (st == NULL)
        {
            printf("ERROR (mon_generate_by_degree).  Memory allocation failed.\n");
            exit(-1);
        }
        
        MON_INIT(st[0]);
        *len = 1;
        return st;
    }
    
    N = _binom_(n + d, d);
    
    st   = (monomial_t *) malloc(N * sizeof(monomial_t));
    sums = (uint32_t *) malloc(N * sizeof(uint32_t));
    inds = (uint32_t *) malloc(N * sizeof(uint32_t));
    if (st == NULL | sums == NULL | inds == NULL)
    {
        printf("ERROR (mon_generate_by_degree).  Memory allocation failed.\n");
        exit(-1);
    }
    
    for (i = 0; i < N; i++)
        MON_INIT(st[i]);
    
    MON_INIT(m);
    
    i = 0;
    sums[0] = 0;
    inds[0] = 0;
    j = N;
    
    while (i >= 0)
    {
        MON_SET(m, st[i]);
        ind = inds[i];
        sum = sums[i];
        i--;
        
        if (ind < n)
        {
            for (exp = 0; exp <= d - sum; exp++)
            {
                i++;
                MON_SET(st[i], m);
                MON_INC_EXP(st[i], ind, exp);
                inds[i] = ind + 1;
                sums[i] = sum + exp;
            }
        }
        else
        {
            j--;
            MON_SET(st[j], m);
            MON_INC_EXP(st[j], ind, d-sum);
        }
    }
    
    MON_CLEAR(m);
    free(sums);
    free(inds);
    
    if (N > 1)
    {
        j = N - 1;
        monomial_t temp;
        for (i = 0; i < N/2; i++)
        {
            MON_SWAP(st[i], st[j]);
            j--;
        }
    }
    
    *len = N;
    return st;
}

/**
 * Generates all monomials in <tt>n + 1</tt> variables of total degree \c d.
 *
 * This method returns an array of length \f$\binom{n+d}{d}\f$, which is also 
 * the value of <tt>*len</tt> on exit.  The monomials are arrange in ascending 
 * inverse lexicographical order.
 */
monomial_t * mon_generate_by_degree_invlex(uint32_t * len, const uint8_t n, 
                                           const uint32_t d)
{
    monomial_t * st;  // Stack (from the bottom), list (from the top)
    uint32_t * sums;
    uint32_t * inds;
    int32_t i;        // Pointer to the top of the stack
    int32_t j;        // Pointer to the bottom of the list
    uint32_t N;       // $\binom{n+d}{d}$
    monomial_t m;
    uint32_t sum, ind;
    exponent_t exp;
    
    // Special case d == 0
    if (d == 0)
    {
        st = (monomial_t *) malloc(sizeof(monomial_t));
        if (st == NULL)
        {
            printf("ERROR (mon_generate_by_degree_invlex).  Memory allocation failed.\n");
            exit(-1);
        }
        
        MON_INIT(st[0]);
        *len = 1;
        return st;
    }
    
    N = _binom_(n + d, d);
    
    st   = (monomial_t *) malloc(N * sizeof(monomial_t));
    sums = (uint32_t *) malloc(N * sizeof(uint32_t));
    inds = (uint32_t *) malloc(N * sizeof(uint32_t));
    if (st == NULL | sums == NULL | inds == NULL)
    {
        printf("ERROR (mon_generate_by_degree_invlex).  Memory allocation failed.\n");
        exit(-1);
    }
    
    for (i = 0; i < N; i++)
        MON_INIT(st[i]);
    
    MON_INIT(m);
    
    i = 0;
    sums[0] = 0;
    inds[0] = n;
    j = N;
    
    while (i >= 0)
    {
        MON_SET(m, st[i]);
        ind = inds[i];
        sum = sums[i];
        i--;
        
        if (ind > 0)
        {
            for (exp = 0; exp <= d - sum; exp++)
            {
                i++;
                MON_SET(st[i], m);
                MON_INC_EXP(st[i], ind, exp);
                inds[i] = ind - 1;
                sums[i] = sum + exp;
            }
        }
        else
        {
            j--;
            MON_SET(st[j], m);
            MON_INC_EXP(st[j], ind, d-sum);
        }
    }
    
    MON_CLEAR(m);
    free(sums);
    free(inds);
    
    *len = N;
    return st;
}


