///////////////////////////////////////////////////////////////////////////////
// Copyright (C) 2009 Sebastian Pancratz                                     //
//                                                                           //
// Distributed under the terms of the GNU General Public License as          //
// published by the Free Software Foundation; either version 2 of the        //
// License, or (at your option) any later version.                           //
//                                                                           //
// http://www.gnu.org/licenses/                                              //
///////////////////////////////////////////////////////////////////////////////

#ifndef _MONOMIAL_H_
#define _MONOMIAL_H_

/**
 * \file    monomial.h
 * \brief   Interface for monomials
 * \author  Sebastian Pancratz
 * \date    Oct--Nov 2009
 */

#include <inttypes.h>

/**
 * \defgroup  Initialization   Initialization and memory management
 * \defgroup  Access           Basic access
 * \defgroup  Comparison       Comparison
 * \defgroup  Multiplication   Multiplication and division
 * \defgroup  Strings          String conversions and I/O
 * \defgroup  Misc             Further methods
 */

///////////////////////////////////////////////////////////////////////////////
// Data types                                                                //
///////////////////////////////////////////////////////////////////////////////

/**
 * \typedef monomial_t
 *
 * Data type for a monomial.
 *
 * We represent a monomial as a 64-bit word.  More precisely, we store the 
 * monomial \f$(i_0, \dotsc, i_n)\f$ using #MON_BITS_PER_EXP bits per 
 * exponent, where the least significant block is used for \f$i_0\f$.  If 
 * \f$n+1\f$ does not divide 64 exactly then the \f$64 \bmod (n+1)\f$ most 
 * significant bits are always \c 0.
 */
typedef uint64_t monomial_t;

/**
 * \typedef exponent_t
 *
 * Data type for an exponent.
 *
 * We represent an exponent as an 8-bit word.  This means that we only 
 * support monomials in the range \f$[0, 256)\f$.
 */
typedef uint8_t exponent_t;

///////////////////////////////////////////////////////////////////////////////
// Object-like macros                                                        //
///////////////////////////////////////////////////////////////////////////////

/**
 * \def MON_BITS_PER_EXP
 *
 * Number of bits per exponent in the word used for the monomial.
 */
#define MON_BITS_PER_EXP  8

/**
 * \def MON_BITMASK_BLOCK
 *
 * Bitmask with all <tt>0</tt>s apart from the MON_BITS_PER_EXP least 
 * significant bits.
 */
#define MON_BITMASK_BLOCK  0x00000000000000FFULL

/**
 * \def MON_BITMASK_ALL
 *
 * Bitmask with all <tt>1</tt>s.
 */
#define MON_BITMASK_ALL  0xFFFFFFFFFFFFFFFFULL

/**
 * \def MON_VARS
 * 
 * String literal of length <tt>n + 1</tt> with default variable names.
 */
#define MON_VARS(n)  (n == 1 ? "XY" : n == 2 ? "XYZ" : n == 3 ? "WXYZ" : n == 4 ? "VWXYZ" : n == 5 ? "UVWXYZ" : n == 6 ? "TUVWXYZ" : "STUVWXYZ")

///////////////////////////////////////////////////////////////////////////////
// Function-like macros                                                      //
///////////////////////////////////////////////////////////////////////////////

///////////////////////////////////////////////////////////////////////////////
// Initialization

/**
 * \addtogroup Initialization
 * @{
 */

/**
 * \def MON_INIT
 *
 * Initializes the monomial to one.
 * 
 * \note  This macro evaluates each argument only once.
 */
#define MON_INIT(x)  ((x) = 0)

/**
 * \def MON_CLEAR
 *
 * Clears all data allocated for this monomial.
 */
#define MON_CLEAR(x)

/**
 * \def MON_SET
 *
 * Sets the monomial \c x to the same value as \c y.
 * 
 * \note  This macro evaluates each argument only once.
 */
#define MON_SET(x, y)  ((x) = (y))

/**
 * \def MON_SWAP
 *
 * Swaps the values of the monomials \c x and \c y.
 * 
 * \warning  This macro possibly evaluates its arguments multiple times.
 */
#define MON_SWAP(x, y)  { monomial_t _t_ = (x); (x) = (y); (y) = _t_; }

/**
 * \def MON_ONE
 *
 * Sets \c x to the monomial one.
 *
 * \note  This macro evaluates each argument only once.
 */
#define MON_ONE(x)  ((x) = 0)

/**
 * @}
 */

///////////////////////////////////////////////////////////////////////////////
// Access

/**
 * \addtogroup Access
 *
 * @{
 */

/**
 * \def MON_GET_EXP
 *
 * Returns the <tt>i</tt>th exponent of the monomial \c x.
 * 
 * \note  This macro evaluates each argument only once.
 */
#define MON_GET_EXP(x, i)  ((exponent_t) (((x) >> ((i) * MON_BITS_PER_EXP)) & MON_BITMASK_BLOCK))

/**
 * \def MON_SET_EXP
 *
 * Sets the <tt>i</tt>th exponent of the monomial \c x to the value \c e.
 * 
 * \warning  This macro possibly evaluates its arguments multiple times.
 */
#define MON_SET_EXP(x, i, e)  ((x) = (((x) & (MON_BITMASK_ALL - (MON_BITMASK_BLOCK << ((i) * MON_BITS_PER_EXP)))) | (((monomial_t)(e)) << ((i) * MON_BITS_PER_EXP))))

/**
 * \def MON_INC_EXP
 *
 * Increments the current value of the <tt>i</tt>th exponent of the 
 * monomial \c x by \c e.
 * 
 * \warning  This macro possibly evaluates its arguments multiple times.
 */
#define MON_INC_EXP(x, i, e)  ((x) = ((x) + (((monomial_t)(e)) << ((i) * MON_BITS_PER_EXP))))

/**
 * \def MON_DEC_EXP
 *
 * Decrements the current value of the <tt>i</tt>th exponent of the 
 * monomial \c x by \c e.
 * 
 * \warning  This macro possibly evaluates its arguments multiple times.
 */
#define MON_DEC_EXP(x, i, e)  ((x) = ((x) - (((monomial_t)(e)) << ((i) * MON_BITS_PER_EXP))))

/**
 * @}
 */

///////////////////////////////////////////////////////////////////////////////
// Comparison

/**
 * \addtogroup Comparison
 * @{
 */

/**
 * \def MON_CMP_INVLEX
 * 
 * Compares the two monomials \c x and \c y inverse-lexicographically.
 * 
 * \warning  This macro possibly evaluates its arguments multiple times.
 */
#define MON_CMP_INVLEX(x, y)  (((x) < (y)) ? -1 : ((x) > (y) ? 1 : 0))

/**
 * \def MON_IS_ONE
 *
 * Returns whether the monomial \c x is one.
 * 
 * \note  This macro evaluates each argument only once.
 */
#define MON_IS_ONE(x)  ((x) == 0)

/**
 * @}
 */

///////////////////////////////////////////////////////////////////////////////
// Multiplication and division

/**
 * \addtogroup Multiplication
 * @{
 */

/**
 * \def MON_MUL
 *
 * Sets \c x to the product of \c y and \c z.
 * 
 * \note  This macro evaluates each argument only once.
 */
#define MON_MUL(x, y, z)  ((x) = (y) + (z))

/**
 * \def MON_DIV
 *
 * Sets \c x to the quotient of \c y and \c z, assuming that \c z divides \c y.
 *
 * \note  This macro evaluates each argument only once.
 */
#define MON_DIV(x, y, z)  ((x) = (y) - (z))

/**
 * @}
 */

///////////////////////////////////////////////////////////////////////////////
// Function declarations                                                     //
///////////////////////////////////////////////////////////////////////////////

/**
 * \addtogroup Access
 * @{
 */

uint32_t mon_degree(monomial_t op);

/**
 * @}
 */

/**
 * \addtogroup Multiplication
 * @{
 */

int32_t mon_divides(monomial_t x, monomial_t y);

/**
 * @}
 */

/**
 * \addtogroup Strings
 * 
 * The format used for a monomial \c m in \c n variables is 
 * <tt>n  m[0] ... m[n-1]</tt>, where there are \e two spaces after the first 
 * integer and single spaces only otherwise.
 *
 * @{
 */

char * mon_to_string(monomial_t x, const uint8_t n);
char * mon_to_string_pretty(monomial_t x, const uint8_t n, const char * vars);
monomial_t mon_from_string(char * str);

/**
 * @}
 */

/**
 * \addtogroup Misc
 * @{
 */

monomial_t * mon_generate_by_degree(uint32_t * len, const uint8_t n, 
                                    const uint32_t d);
monomial_t * mon_generate_by_degree_invlex(uint32_t * len, const uint8_t n, 
                                           const uint32_t d);

/**
 * @}
 */

#endif

